using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Pixelplacement;

public class OwnerPetPair {
    public CharacterComponent owner;
    public CharacterComponent pet;
}

[System.Serializable]
public class Card {
    public CardAsset asset;
    public int durability;
}

public class GameManager : Singleton<GameManager>
{
    
    public GeneralCardData generalCardData;

    [SerializeField] private Vector2 tileSize;

    [SerializeField] private List<Card> hand;

    [SerializeField] private RectTransform handPanel;

    [SerializeField] private GameObject tileUIPrefab;

    [SerializeField] private float cardSelectionTweenTime = 0.1f, cardSelectionOffset = 8.0f;

    private Dictionary<Vector2Int, TileComponent> tiles;
    private List<CharacterComponent> characters;
    private Dictionary<int, OwnerPetPair> pairings;
    private List<CharacterComponent> killers;
    private bool isRunning, isFastForward, levelFailed, levelWon;
    public float timePerStep = 0.5f, fastForwardSpeedUp = 5.0f;
    private float timeTillNextStep = Mathf.Infinity;

    private List<TileUIComponent> tilesInHand;
    private TileUIComponent selected;


    public bool IsRunning { get { return isRunning; } }

    private Vector3 getWorldPositionFromGridPosition(Vector2Int gridPosition){
        return new Vector3(gridPosition.x * tileSize.x, gridPosition.y * tileSize.y, 0);
    }

    private Vector2Int getGridPositionFromWorldPosition(Vector3 worldPosition){
        return new Vector2Int(Mathf.RoundToInt(worldPosition.x/tileSize.x), Mathf.RoundToInt(worldPosition.y/tileSize.y));
    } 

    // Start is called before the first frame update
    void Start()
    {
        tiles = new Dictionary<Vector2Int, TileComponent>();
        foreach(TileComponent tile in FindObjectsOfType<TileComponent>()){
            Transform tileTransform = tile.GetComponent<Transform>();
            Vector3 worldPosition = tileTransform.position;
            Vector2Int gridPosition = getGridPositionFromWorldPosition(worldPosition);
            #if UNITY_EDITOR
            if(tiles.ContainsKey(gridPosition)) Debug.LogError("Tiles overlap at " + gridPosition.ToString());
            #endif
            tiles.Add(gridPosition, tile);
        }
        #if UNITY_EDITOR
        Debug.Log(tiles.Count.ToString() + " tiles found");
        #endif
        characters = new List<CharacterComponent>();
        pairings = new Dictionary<int, OwnerPetPair>();
        killers = new List<CharacterComponent>();
        foreach(CharacterComponent character in FindObjectsOfType<CharacterComponent>()){
            Transform characterTransform = character.GetComponent<Transform>();
            Vector3 worldPosition = characterTransform.position;
            Vector2Int gridPosition = getGridPositionFromWorldPosition(worldPosition);
            character.originalGridPosition = gridPosition;
            character.PrepareForRun();
            character.UpdateVisuals();
            characters.Add(character);
            if(character.characterType == CharacterType.Death){
                killers.Add(character);
            } else if(pairings.TryGetValue(character.matchID, out OwnerPetPair pair)){
                if(character.characterType == CharacterType.Owner) pair.owner = character;
                else pair.pet = character;    
            } else {
                pair = new OwnerPetPair();
                if(character.characterType == CharacterType.Owner) pair.owner = character;
                else pair.pet = character;
                pairings.Add(character.matchID, pair);
            }
        }
        tilesInHand =  new List<TileUIComponent>();
        foreach(Card card in hand){
            GameObject cardGameObject = GameObject.Instantiate(tileUIPrefab, Vector2.zero, Quaternion.identity);
            cardGameObject.GetComponent<RectTransform>().SetParent(handPanel, false);
            TileUIComponent tileUI = cardGameObject.GetComponent<TileUIComponent>();
            tileUI.AssignCard(card);
            tilesInHand.Add(tileUI);
        }
    }

    public CharacterComponent GetOther(CharacterComponent character){
        OwnerPetPair pair = pairings[character.matchID];
        if(character.characterType == CharacterType.Owner) return pair.pet;
        else return pair.owner;
    }

    // Update is called once per frame
    void Update()
    {
        if(Input.GetMouseButtonDown(1)){
            DeselectTileFromHand();
        }
        if(isRunning && !levelFailed){
            if(timeTillNextStep <= 0){
                timeTillNextStep = timePerStep;
                if(isFastForward) timeTillNextStep /= fastForwardSpeedUp;

                foreach(CharacterComponent killer in killers){
                    foreach(CharacterComponent character in characters){
                        if(character.characterType == CharacterType.Death) continue;
                        if(killer.currentGridPosition == character.currentGridPosition){
                            //TODO KABOOM
                            levelFailed = true;
                        }
                    }
                }

                if(!levelFailed) foreach(OwnerPetPair pairing in pairings.Values){
                    // Commented for now since I still don't have all the pairs
                    //if(pairing.owner.currentGridPosition == pairing.pet.currentGridPosition){
                        //TODO Win
                    //}
                }

                if(!levelFailed) foreach(CharacterComponent character in characters){
                    TileComponent tile = tiles[character.currentGridPosition];
                    if(tile.cardBehaviour != null){
                        tile.cardBehaviour.Apply(tile, character);
                    } else if(tile.tileType == TileType.Death){
                        // TODO
                        if(character.characterType != CharacterType.Death) levelFailed = true;
                    }
                }

                if(!levelFailed) foreach(CharacterComponent character in characters){
                    Vector2Int nextGridPosition = character.currentGridPosition + character.currentDirection.ToVector();
                    if(tiles.TryGetValue(nextGridPosition, out TileComponent nextTile)){
                        if(nextTile.tileType != TileType.Wall && !nextTile.destroyed){
                            character.currentGridPosition = nextGridPosition;
                            Vector3 worldPosition = getWorldPositionFromGridPosition(nextGridPosition);
                            Tween.Position(character.transform, worldPosition, timeTillNextStep * 0.5f, 0, Tween.EaseInOut);
                        }
                    }
                    character.UpdateVisuals();
                }

                
                
            } else {
                timeTillNextStep -= Time.deltaTime;
            }
        }
    }

    public void Run(){
        DeselectTileFromHand();
        isRunning = true;
        timeTillNextStep = timePerStep;
        foreach(CharacterComponent character in characters){
            character.PrepareForRun();
        }
        foreach(TileComponent tile in tiles.Values){
            tile.UpdateTileVisuals();
        }
    }

    public void Restart(){
        Tween.StopAll();
    }

    public void DeselectTileFromHand(){
        if(selected != null){
            Tween.AnchoredPosition(selected.cardGroup, new Vector2(0, 0), cardSelectionTweenTime, 0, Tween.EaseInOutStrong);
        }
    }

    public void SelectTileFromHand(TileUIComponent tile){
        DeselectTileFromHand();
        selected = tile;
        Tween.AnchoredPosition(selected.cardGroup, new Vector2(0, cardSelectionOffset), cardSelectionTweenTime, 0, Tween.EaseInOutStrong);
    }

    private void ReturnCardToHand(Card card){
        foreach(TileUIComponent tileUI in tilesInHand){
            if(tileUI.card == card) tileUI.gameObject.SetActive(true);
        }
    }

    public void PutCardOnTile(TileComponent tile){
        if(selected != null){
            if(tile.card != null){
                ReturnCardToHand(tile.card);
                tile.ClearCard();
            }
            tile.AssignCard(selected.card);
            tile.UpdateTileVisuals();
            DeselectTileFromHand();
            selected.gameObject.SetActive(false);
        }
    }

    public void ClearCardOnTile(TileComponent tile){
        if(tile.card != null){
            ReturnCardToHand(tile.card);
            tile.ClearCard();
            tile.UpdateTileVisuals();
        }
    }

    public void BackToMenu(){

    }
}
